(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('pbf')) :
    typeof define === 'function' && define.amd ? define(['pbf'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, (global[''] = global[''] || {}, global['']['/dist/mapbox-gl-arcgis-featureserver'] = global['']['/dist/mapbox-gl-arcgis-featureserver'] || {}, global['']['/dist/mapbox-gl-arcgis-featureserver'].js = factory(global.Pbf)));
}(this, (function (Pbf) { 'use strict';

    function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

    var Pbf__default = /*#__PURE__*/_interopDefaultLegacy(Pbf);

    var d2r = Math.PI / 180,
        r2d = 180 / Math.PI;

    /**
     * Get the bbox of a tile
     *
     * @name tileToBBOX
     * @param {Array<number>} tile
     * @returns {Array<number>} bbox
     * @example
     * var bbox = tileToBBOX([5, 10, 10])
     * //=bbox
     */
    function tileToBBOX(tile) {
        var e = tile2lon(tile[0] + 1, tile[2]);
        var w = tile2lon(tile[0], tile[2]);
        var s = tile2lat(tile[1] + 1, tile[2]);
        var n = tile2lat(tile[1], tile[2]);
        return [w, s, e, n];
    }

    /**
     * Get a geojson representation of a tile
     *
     * @name tileToGeoJSON
     * @param {Array<number>} tile
     * @returns {Feature<Polygon>}
     * @example
     * var poly = tileToGeoJSON([5, 10, 10])
     * //=poly
     */
    function tileToGeoJSON(tile) {
        var bbox = tileToBBOX(tile);
        var poly = {
            type: 'Polygon',
            coordinates: [[
                [bbox[0], bbox[3]],
                [bbox[0], bbox[1]],
                [bbox[2], bbox[1]],
                [bbox[2], bbox[3]],
                [bbox[0], bbox[3]]
            ]]
        };
        return poly;
    }

    function tile2lon(x, z) {
        return x / Math.pow(2, z) * 360 - 180;
    }

    function tile2lat(y, z) {
        var n = Math.PI - 2 * Math.PI * y / Math.pow(2, z);
        return r2d * Math.atan(0.5 * (Math.exp(n) - Math.exp(-n)));
    }

    /**
     * Get the tile for a point at a specified zoom level
     *
     * @name pointToTile
     * @param {number} lon
     * @param {number} lat
     * @param {number} z
     * @returns {Array<number>} tile
     * @example
     * var tile = pointToTile(1, 1, 20)
     * //=tile
     */
    function pointToTile(lon, lat, z) {
        var tile = pointToTileFraction(lon, lat, z);
        tile[0] = Math.floor(tile[0]);
        tile[1] = Math.floor(tile[1]);
        return tile;
    }

    /**
     * Get the 4 tiles one zoom level higher
     *
     * @name getChildren
     * @param {Array<number>} tile
     * @returns {Array<Array<number>>} tiles
     * @example
     * var tiles = getChildren([5, 10, 10])
     * //=tiles
     */
    function getChildren(tile) {
        return [
            [tile[0] * 2, tile[1] * 2, tile[2] + 1],
            [tile[0] * 2 + 1, tile[1] * 2, tile[2 ] + 1],
            [tile[0] * 2 + 1, tile[1] * 2 + 1, tile[2] + 1],
            [tile[0] * 2, tile[1] * 2 + 1, tile[2] + 1]
        ];
    }

    /**
     * Get the tile one zoom level lower
     *
     * @name getParent
     * @param {Array<number>} tile
     * @returns {Array<number>} tile
     * @example
     * var tile = getParent([5, 10, 10])
     * //=tile
     */
    function getParent(tile) {
        return [tile[0] >> 1, tile[1] >> 1, tile[2] - 1];
    }

    function getSiblings(tile) {
        return getChildren(getParent(tile));
    }

    /**
     * Get the 3 sibling tiles for a tile
     *
     * @name getSiblings
     * @param {Array<number>} tile
     * @returns {Array<Array<number>>} tiles
     * @example
     * var tiles = getSiblings([5, 10, 10])
     * //=tiles
     */
    function hasSiblings(tile, tiles) {
        var siblings = getSiblings(tile);
        for (var i = 0; i < siblings.length; i++) {
            if (!hasTile(tiles, siblings[i])) return false;
        }
        return true;
    }

    /**
     * Check to see if an array of tiles contains a particular tile
     *
     * @name hasTile
     * @param {Array<Array<number>>} tiles
     * @param {Array<number>} tile
     * @returns {boolean}
     * @example
     * var tiles = [
     *     [0, 0, 5],
     *     [0, 1, 5],
     *     [1, 1, 5],
     *     [1, 0, 5]
     * ]
     * hasTile(tiles, [0, 0, 5])
     * //=boolean
     */
    function hasTile(tiles, tile) {
        for (var i = 0; i < tiles.length; i++) {
            if (tilesEqual(tiles[i], tile)) return true;
        }
        return false;
    }

    /**
     * Check to see if two tiles are the same
     *
     * @name tilesEqual
     * @param {Array<number>} tile1
     * @param {Array<number>} tile2
     * @returns {boolean}
     * @example
     * tilesEqual([0, 1, 5], [0, 0, 5])
     * //=boolean
     */
    function tilesEqual(tile1, tile2) {
        return (
            tile1[0] === tile2[0] &&
            tile1[1] === tile2[1] &&
            tile1[2] === tile2[2]
        );
    }

    /**
     * Get the quadkey for a tile
     *
     * @name tileToQuadkey
     * @param {Array<number>} tile
     * @returns {string} quadkey
     * @example
     * var quadkey = tileToQuadkey([0, 1, 5])
     * //=quadkey
     */
    function tileToQuadkey(tile) {
        var index = '';
        for (var z = tile[2]; z > 0; z--) {
            var b = 0;
            var mask = 1 << (z - 1);
            if ((tile[0] & mask) !== 0) b++;
            if ((tile[1] & mask) !== 0) b += 2;
            index += b.toString();
        }
        return index;
    }

    /**
     * Get the tile for a quadkey
     *
     * @name quadkeyToTile
     * @param {string} quadkey
     * @returns {Array<number>} tile
     * @example
     * var tile = quadkeyToTile('00001033')
     * //=tile
     */
    function quadkeyToTile(quadkey) {
        var x = 0;
        var y = 0;
        var z = quadkey.length;

        for (var i = z; i > 0; i--) {
            var mask = 1 << (i - 1);
            var q = +quadkey[z - i];
            if (q === 1) x |= mask;
            if (q === 2) y |= mask;
            if (q === 3) {
                x |= mask;
                y |= mask;
            }
        }
        return [x, y, z];
    }

    /**
     * Get the smallest tile to cover a bbox
     *
     * @name bboxToTile
     * @param {Array<number>} bbox
     * @returns {Array<number>} tile
     * @example
     * var tile = bboxToTile([ -178, 84, -177, 85 ])
     * //=tile
     */
    function bboxToTile(bboxCoords) {
        var min = pointToTile(bboxCoords[0], bboxCoords[1], 32);
        var max = pointToTile(bboxCoords[2], bboxCoords[3], 32);
        var bbox = [min[0], min[1], max[0], max[1]];

        var z = getBboxZoom(bbox);
        if (z === 0) return [0, 0, 0];
        var x = bbox[0] >>> (32 - z);
        var y = bbox[1] >>> (32 - z);
        return [x, y, z];
    }

    function getBboxZoom(bbox) {
        var MAX_ZOOM = 28;
        for (var z = 0; z < MAX_ZOOM; z++) {
            var mask = 1 << (32 - (z + 1));
            if (((bbox[0] & mask) !== (bbox[2] & mask)) ||
                ((bbox[1] & mask) !== (bbox[3] & mask))) {
                return z;
            }
        }

        return MAX_ZOOM;
    }

    /**
     * Get the precise fractional tile location for a point at a zoom level
     *
     * @name pointToTileFraction
     * @param {number} lon
     * @param {number} lat
     * @param {number} z
     * @returns {Array<number>} tile fraction
     * var tile = pointToTileFraction(30.5, 50.5, 15)
     * //=tile
     */
    function pointToTileFraction(lon, lat, z) {
        var sin = Math.sin(lat * d2r),
            z2 = Math.pow(2, z),
            x = z2 * (lon / 360 + 0.5),
            y = z2 * (0.5 - 0.25 * Math.log((1 + sin) / (1 - sin)) / Math.PI);

        // Wrap Tile X
        x = x % z2;
        if (x < 0) x = x + z2;
        return [x, y, z];
    }

    var tilebelt = {
        tileToGeoJSON: tileToGeoJSON,
        tileToBBOX: tileToBBOX,
        getChildren: getChildren,
        getParent: getParent,
        getSiblings: getSiblings,
        hasTile: hasTile,
        hasSiblings: hasSiblings,
        tilesEqual: tilesEqual,
        tileToQuadkey: tileToQuadkey,
        quadkeyToTile: quadkeyToTile,
        pointToTile: pointToTile,
        bboxToTile: bboxToTile,
        pointToTileFraction: pointToTileFraction
    };

    var FeatureCollectionPBuffer_1;

    // FeatureCollectionPBuffer ========================================

    var FeatureCollectionPBuffer = FeatureCollectionPBuffer_1 = {};

    FeatureCollectionPBuffer.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer._readField, {version: "", queryResult: null}, end);
    };
    FeatureCollectionPBuffer._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.version = pbf.readString();
        else if (tag === 2) obj.queryResult = FeatureCollectionPBuffer.QueryResult.read(pbf, pbf.readVarint() + pbf.pos);
    };
    FeatureCollectionPBuffer.write = function (obj, pbf) {
        if (obj.version) pbf.writeStringField(1, obj.version);
        if (obj.queryResult) pbf.writeMessage(2, FeatureCollectionPBuffer.QueryResult.write, obj.queryResult);
    };

    FeatureCollectionPBuffer.GeometryType = {
        "esriGeometryTypePoint": {
            "value": 0,
            "options": {}
        },
        "esriGeometryTypeMultipoint": {
            "value": 1,
            "options": {}
        },
        "esriGeometryTypePolyline": {
            "value": 2,
            "options": {}
        },
        "esriGeometryTypePolygon": {
            "value": 3,
            "options": {}
        },
        "esriGeometryTypeMultipatch": {
            "value": 4,
            "options": {}
        },
        "esriGeometryTypeNone": {
            "value": 127,
            "options": {}
        }
    };

    FeatureCollectionPBuffer.FieldType = {
        "esriFieldTypeSmallInteger": {
            "value": 0,
            "options": {}
        },
        "esriFieldTypeInteger": {
            "value": 1,
            "options": {}
        },
        "esriFieldTypeSingle": {
            "value": 2,
            "options": {}
        },
        "esriFieldTypeDouble": {
            "value": 3,
            "options": {}
        },
        "esriFieldTypeString": {
            "value": 4,
            "options": {}
        },
        "esriFieldTypeDate": {
            "value": 5,
            "options": {}
        },
        "esriFieldTypeOID": {
            "value": 6,
            "options": {}
        },
        "esriFieldTypeGeometry": {
            "value": 7,
            "options": {}
        },
        "esriFieldTypeBlob": {
            "value": 8,
            "options": {}
        },
        "esriFieldTypeRaster": {
            "value": 9,
            "options": {}
        },
        "esriFieldTypeGUID": {
            "value": 10,
            "options": {}
        },
        "esriFieldTypeGlobalID": {
            "value": 11,
            "options": {}
        },
        "esriFieldTypeXML": {
            "value": 12,
            "options": {}
        }
    };

    FeatureCollectionPBuffer.SQLType = {
        "sqlTypeBigInt": {
            "value": 0,
            "options": {}
        },
        "sqlTypeBinary": {
            "value": 1,
            "options": {}
        },
        "sqlTypeBit": {
            "value": 2,
            "options": {}
        },
        "sqlTypeChar": {
            "value": 3,
            "options": {}
        },
        "sqlTypeDate": {
            "value": 4,
            "options": {}
        },
        "sqlTypeDecimal": {
            "value": 5,
            "options": {}
        },
        "sqlTypeDouble": {
            "value": 6,
            "options": {}
        },
        "sqlTypeFloat": {
            "value": 7,
            "options": {}
        },
        "sqlTypeGeometry": {
            "value": 8,
            "options": {}
        },
        "sqlTypeGUID": {
            "value": 9,
            "options": {}
        },
        "sqlTypeInteger": {
            "value": 10,
            "options": {}
        },
        "sqlTypeLongNVarchar": {
            "value": 11,
            "options": {}
        },
        "sqlTypeLongVarbinary": {
            "value": 12,
            "options": {}
        },
        "sqlTypeLongVarchar": {
            "value": 13,
            "options": {}
        },
        "sqlTypeNChar": {
            "value": 14,
            "options": {}
        },
        "sqlTypeNVarchar": {
            "value": 15,
            "options": {}
        },
        "sqlTypeOther": {
            "value": 16,
            "options": {}
        },
        "sqlTypeReal": {
            "value": 17,
            "options": {}
        },
        "sqlTypeSmallInt": {
            "value": 18,
            "options": {}
        },
        "sqlTypeSqlXml": {
            "value": 19,
            "options": {}
        },
        "sqlTypeTime": {
            "value": 20,
            "options": {}
        },
        "sqlTypeTimestamp": {
            "value": 21,
            "options": {}
        },
        "sqlTypeTimestamp2": {
            "value": 22,
            "options": {}
        },
        "sqlTypeTinyInt": {
            "value": 23,
            "options": {}
        },
        "sqlTypeVarbinary": {
            "value": 24,
            "options": {}
        },
        "sqlTypeVarchar": {
            "value": 25,
            "options": {}
        }
    };

    FeatureCollectionPBuffer.QuantizeOriginPostion = {
        "upperLeft": {
            "value": 0,
            "options": {}
        },
        "lowerLeft": {
            "value": 1,
            "options": {}
        }
    };

    // FeatureCollectionPBuffer.SpatialReference ========================================

    FeatureCollectionPBuffer.SpatialReference = {};

    FeatureCollectionPBuffer.SpatialReference.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.SpatialReference._readField, {wkid: 0, lastestWkid: 0, vcsWkid: 0, latestVcsWkid: 0, wkt: ""}, end);
    };
    FeatureCollectionPBuffer.SpatialReference._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.wkid = pbf.readVarint();
        else if (tag === 2) obj.lastestWkid = pbf.readVarint();
        else if (tag === 3) obj.vcsWkid = pbf.readVarint();
        else if (tag === 4) obj.latestVcsWkid = pbf.readVarint();
        else if (tag === 5) obj.wkt = pbf.readString();
    };
    FeatureCollectionPBuffer.SpatialReference.write = function (obj, pbf) {
        if (obj.wkid) pbf.writeVarintField(1, obj.wkid);
        if (obj.lastestWkid) pbf.writeVarintField(2, obj.lastestWkid);
        if (obj.vcsWkid) pbf.writeVarintField(3, obj.vcsWkid);
        if (obj.latestVcsWkid) pbf.writeVarintField(4, obj.latestVcsWkid);
        if (obj.wkt) pbf.writeStringField(5, obj.wkt);
    };

    // FeatureCollectionPBuffer.Field ========================================

    FeatureCollectionPBuffer.Field = {};

    FeatureCollectionPBuffer.Field.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Field._readField, {name: "", fieldType: 0, alias: "", sqlType: 0, domain: "", defaultValue: ""}, end);
    };
    FeatureCollectionPBuffer.Field._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.name = pbf.readString();
        else if (tag === 2) obj.fieldType = pbf.readVarint();
        else if (tag === 3) obj.alias = pbf.readString();
        else if (tag === 4) obj.sqlType = pbf.readVarint();
        else if (tag === 5) obj.domain = pbf.readString();
        else if (tag === 6) obj.defaultValue = pbf.readString();
    };
    FeatureCollectionPBuffer.Field.write = function (obj, pbf) {
        if (obj.name) pbf.writeStringField(1, obj.name);
        if (obj.fieldType) pbf.writeVarintField(2, obj.fieldType);
        if (obj.alias) pbf.writeStringField(3, obj.alias);
        if (obj.sqlType) pbf.writeVarintField(4, obj.sqlType);
        if (obj.domain) pbf.writeStringField(5, obj.domain);
        if (obj.defaultValue) pbf.writeStringField(6, obj.defaultValue);
    };

    // FeatureCollectionPBuffer.Value ========================================

    FeatureCollectionPBuffer.Value = {};

    FeatureCollectionPBuffer.Value.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Value._readField, {string_value: "", value_type: null, float_value: 0, double_value: 0, sint_value: 0, uint_value: 0, int64_value: 0, uint64_value: 0, sint64_value: 0, bool_value: false}, end);
    };
    FeatureCollectionPBuffer.Value._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.string_value = pbf.readString(), obj.value_type = "string_value";
        else if (tag === 2) obj.float_value = pbf.readFloat(), obj.value_type = "float_value";
        else if (tag === 3) obj.double_value = pbf.readDouble(), obj.value_type = "double_value";
        else if (tag === 4) obj.sint_value = pbf.readSVarint(), obj.value_type = "sint_value";
        else if (tag === 5) obj.uint_value = pbf.readVarint(), obj.value_type = "uint_value";
        else if (tag === 6) obj.int64_value = pbf.readVarint(true), obj.value_type = "int64_value";
        else if (tag === 7) obj.uint64_value = pbf.readVarint(), obj.value_type = "uint64_value";
        else if (tag === 8) obj.sint64_value = pbf.readSVarint(), obj.value_type = "sint64_value";
        else if (tag === 9) obj.bool_value = pbf.readBoolean(), obj.value_type = "bool_value";
    };
    FeatureCollectionPBuffer.Value.write = function (obj, pbf) {
        if (obj.string_value) pbf.writeStringField(1, obj.string_value);
        if (obj.float_value) pbf.writeFloatField(2, obj.float_value);
        if (obj.double_value) pbf.writeDoubleField(3, obj.double_value);
        if (obj.sint_value) pbf.writeSVarintField(4, obj.sint_value);
        if (obj.uint_value) pbf.writeVarintField(5, obj.uint_value);
        if (obj.int64_value) pbf.writeVarintField(6, obj.int64_value);
        if (obj.uint64_value) pbf.writeVarintField(7, obj.uint64_value);
        if (obj.sint64_value) pbf.writeSVarintField(8, obj.sint64_value);
        if (obj.bool_value) pbf.writeBooleanField(9, obj.bool_value);
    };

    // FeatureCollectionPBuffer.Geometry ========================================

    FeatureCollectionPBuffer.Geometry = {};

    FeatureCollectionPBuffer.Geometry.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Geometry._readField, {lengths: [], coords: []}, end);
    };
    FeatureCollectionPBuffer.Geometry._readField = function (tag, obj, pbf) {
        if (tag === 2) pbf.readPackedVarint(obj.lengths);
        else if (tag === 3) pbf.readPackedSVarint(obj.coords);
    };
    FeatureCollectionPBuffer.Geometry.write = function (obj, pbf) {
        if (obj.lengths) pbf.writePackedVarint(2, obj.lengths);
        if (obj.coords) pbf.writePackedSVarint(3, obj.coords);
    };

    // FeatureCollectionPBuffer.esriShapeBuffer ========================================

    FeatureCollectionPBuffer.esriShapeBuffer = {};

    FeatureCollectionPBuffer.esriShapeBuffer.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.esriShapeBuffer._readField, {bytes: null}, end);
    };
    FeatureCollectionPBuffer.esriShapeBuffer._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.bytes = pbf.readBytes();
    };
    FeatureCollectionPBuffer.esriShapeBuffer.write = function (obj, pbf) {
        if (obj.bytes) pbf.writeBytesField(1, obj.bytes);
    };

    // FeatureCollectionPBuffer.Feature ========================================

    FeatureCollectionPBuffer.Feature = {};

    FeatureCollectionPBuffer.Feature.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Feature._readField, {attributes: [], geometry: null, compressed_geometry: null, shapeBuffer: null, centroid: null}, end);
    };
    FeatureCollectionPBuffer.Feature._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.attributes.push(FeatureCollectionPBuffer.Value.read(pbf, pbf.readVarint() + pbf.pos));
        else if (tag === 2) obj.geometry = FeatureCollectionPBuffer.Geometry.read(pbf, pbf.readVarint() + pbf.pos), obj.compressed_geometry = "geometry";
        else if (tag === 3) obj.shapeBuffer = FeatureCollectionPBuffer.esriShapeBuffer.read(pbf, pbf.readVarint() + pbf.pos), obj.compressed_geometry = "shapeBuffer";
        else if (tag === 4) obj.centroid = FeatureCollectionPBuffer.Geometry.read(pbf, pbf.readVarint() + pbf.pos);
    };
    FeatureCollectionPBuffer.Feature.write = function (obj, pbf) {
        if (obj.attributes) for (var i = 0; i < obj.attributes.length; i++) pbf.writeMessage(1, FeatureCollectionPBuffer.Value.write, obj.attributes[i]);
        if (obj.geometry) pbf.writeMessage(2, FeatureCollectionPBuffer.Geometry.write, obj.geometry);
        if (obj.shapeBuffer) pbf.writeMessage(3, FeatureCollectionPBuffer.esriShapeBuffer.write, obj.shapeBuffer);
        if (obj.centroid) pbf.writeMessage(4, FeatureCollectionPBuffer.Geometry.write, obj.centroid);
    };

    // FeatureCollectionPBuffer.UniqueIdField ========================================

    FeatureCollectionPBuffer.UniqueIdField = {};

    FeatureCollectionPBuffer.UniqueIdField.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.UniqueIdField._readField, {name: "", isSystemMaintained: false}, end);
    };
    FeatureCollectionPBuffer.UniqueIdField._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.name = pbf.readString();
        else if (tag === 2) obj.isSystemMaintained = pbf.readBoolean();
    };
    FeatureCollectionPBuffer.UniqueIdField.write = function (obj, pbf) {
        if (obj.name) pbf.writeStringField(1, obj.name);
        if (obj.isSystemMaintained) pbf.writeBooleanField(2, obj.isSystemMaintained);
    };

    // FeatureCollectionPBuffer.GeometryProperties ========================================

    FeatureCollectionPBuffer.GeometryProperties = {};

    FeatureCollectionPBuffer.GeometryProperties.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.GeometryProperties._readField, {shapeAreaFieldName: "", shapeLengthFieldName: "", units: ""}, end);
    };
    FeatureCollectionPBuffer.GeometryProperties._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.shapeAreaFieldName = pbf.readString();
        else if (tag === 2) obj.shapeLengthFieldName = pbf.readString();
        else if (tag === 3) obj.units = pbf.readString();
    };
    FeatureCollectionPBuffer.GeometryProperties.write = function (obj, pbf) {
        if (obj.shapeAreaFieldName) pbf.writeStringField(1, obj.shapeAreaFieldName);
        if (obj.shapeLengthFieldName) pbf.writeStringField(2, obj.shapeLengthFieldName);
        if (obj.units) pbf.writeStringField(3, obj.units);
    };

    // FeatureCollectionPBuffer.ServerGens ========================================

    FeatureCollectionPBuffer.ServerGens = {};

    FeatureCollectionPBuffer.ServerGens.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.ServerGens._readField, {minServerGen: 0, serverGen: 0}, end);
    };
    FeatureCollectionPBuffer.ServerGens._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.minServerGen = pbf.readVarint();
        else if (tag === 2) obj.serverGen = pbf.readVarint();
    };
    FeatureCollectionPBuffer.ServerGens.write = function (obj, pbf) {
        if (obj.minServerGen) pbf.writeVarintField(1, obj.minServerGen);
        if (obj.serverGen) pbf.writeVarintField(2, obj.serverGen);
    };

    // FeatureCollectionPBuffer.Scale ========================================

    FeatureCollectionPBuffer.Scale = {};

    FeatureCollectionPBuffer.Scale.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Scale._readField, {xScale: 0, yScale: 0, mScale: 0, zScale: 0}, end);
    };
    FeatureCollectionPBuffer.Scale._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.xScale = pbf.readDouble();
        else if (tag === 2) obj.yScale = pbf.readDouble();
        else if (tag === 3) obj.mScale = pbf.readDouble();
        else if (tag === 4) obj.zScale = pbf.readDouble();
    };
    FeatureCollectionPBuffer.Scale.write = function (obj, pbf) {
        if (obj.xScale) pbf.writeDoubleField(1, obj.xScale);
        if (obj.yScale) pbf.writeDoubleField(2, obj.yScale);
        if (obj.mScale) pbf.writeDoubleField(3, obj.mScale);
        if (obj.zScale) pbf.writeDoubleField(4, obj.zScale);
    };

    // FeatureCollectionPBuffer.Translate ========================================

    FeatureCollectionPBuffer.Translate = {};

    FeatureCollectionPBuffer.Translate.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Translate._readField, {xTranslate: 0, yTranslate: 0, mTranslate: 0, zTranslate: 0}, end);
    };
    FeatureCollectionPBuffer.Translate._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.xTranslate = pbf.readDouble();
        else if (tag === 2) obj.yTranslate = pbf.readDouble();
        else if (tag === 3) obj.mTranslate = pbf.readDouble();
        else if (tag === 4) obj.zTranslate = pbf.readDouble();
    };
    FeatureCollectionPBuffer.Translate.write = function (obj, pbf) {
        if (obj.xTranslate) pbf.writeDoubleField(1, obj.xTranslate);
        if (obj.yTranslate) pbf.writeDoubleField(2, obj.yTranslate);
        if (obj.mTranslate) pbf.writeDoubleField(3, obj.mTranslate);
        if (obj.zTranslate) pbf.writeDoubleField(4, obj.zTranslate);
    };

    // FeatureCollectionPBuffer.Transform ========================================

    FeatureCollectionPBuffer.Transform = {};

    FeatureCollectionPBuffer.Transform.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.Transform._readField, {quantizeOriginPostion: 0, scale: null, translate: null}, end);
    };
    FeatureCollectionPBuffer.Transform._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.quantizeOriginPostion = pbf.readVarint();
        else if (tag === 2) obj.scale = FeatureCollectionPBuffer.Scale.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 3) obj.translate = FeatureCollectionPBuffer.Translate.read(pbf, pbf.readVarint() + pbf.pos);
    };
    FeatureCollectionPBuffer.Transform.write = function (obj, pbf) {
        if (obj.quantizeOriginPostion) pbf.writeVarintField(1, obj.quantizeOriginPostion);
        if (obj.scale) pbf.writeMessage(2, FeatureCollectionPBuffer.Scale.write, obj.scale);
        if (obj.translate) pbf.writeMessage(3, FeatureCollectionPBuffer.Translate.write, obj.translate);
    };

    // FeatureCollectionPBuffer.FeatureResult ========================================

    FeatureCollectionPBuffer.FeatureResult = {};

    FeatureCollectionPBuffer.FeatureResult.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.FeatureResult._readField, {objectIdFieldName: "", uniqueIdField: null, globalIdFieldName: "", geohashFieldName: "", geometryProperties: null, serverGens: null, geometryType: 0, spatialReference: null, exceededTransferLimit: false, hasZ: false, hasM: false, transform: null, fields: [], values: [], features: []}, end);
    };
    FeatureCollectionPBuffer.FeatureResult._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.objectIdFieldName = pbf.readString();
        else if (tag === 2) obj.uniqueIdField = FeatureCollectionPBuffer.UniqueIdField.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 3) obj.globalIdFieldName = pbf.readString();
        else if (tag === 4) obj.geohashFieldName = pbf.readString();
        else if (tag === 5) obj.geometryProperties = FeatureCollectionPBuffer.GeometryProperties.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 6) obj.serverGens = FeatureCollectionPBuffer.ServerGens.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 7) obj.geometryType = pbf.readVarint();
        else if (tag === 8) obj.spatialReference = FeatureCollectionPBuffer.SpatialReference.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 9) obj.exceededTransferLimit = pbf.readBoolean();
        else if (tag === 10) obj.hasZ = pbf.readBoolean();
        else if (tag === 11) obj.hasM = pbf.readBoolean();
        else if (tag === 12) obj.transform = FeatureCollectionPBuffer.Transform.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 13) obj.fields.push(FeatureCollectionPBuffer.Field.read(pbf, pbf.readVarint() + pbf.pos));
        else if (tag === 14) obj.values.push(FeatureCollectionPBuffer.Value.read(pbf, pbf.readVarint() + pbf.pos));
        else if (tag === 15) obj.features.push(FeatureCollectionPBuffer.Feature.read(pbf, pbf.readVarint() + pbf.pos));
    };
    FeatureCollectionPBuffer.FeatureResult.write = function (obj, pbf) {
        if (obj.objectIdFieldName) pbf.writeStringField(1, obj.objectIdFieldName);
        if (obj.uniqueIdField) pbf.writeMessage(2, FeatureCollectionPBuffer.UniqueIdField.write, obj.uniqueIdField);
        if (obj.globalIdFieldName) pbf.writeStringField(3, obj.globalIdFieldName);
        if (obj.geohashFieldName) pbf.writeStringField(4, obj.geohashFieldName);
        if (obj.geometryProperties) pbf.writeMessage(5, FeatureCollectionPBuffer.GeometryProperties.write, obj.geometryProperties);
        if (obj.serverGens) pbf.writeMessage(6, FeatureCollectionPBuffer.ServerGens.write, obj.serverGens);
        if (obj.geometryType) pbf.writeVarintField(7, obj.geometryType);
        if (obj.spatialReference) pbf.writeMessage(8, FeatureCollectionPBuffer.SpatialReference.write, obj.spatialReference);
        if (obj.exceededTransferLimit) pbf.writeBooleanField(9, obj.exceededTransferLimit);
        if (obj.hasZ) pbf.writeBooleanField(10, obj.hasZ);
        if (obj.hasM) pbf.writeBooleanField(11, obj.hasM);
        if (obj.transform) pbf.writeMessage(12, FeatureCollectionPBuffer.Transform.write, obj.transform);
        if (obj.fields) for (var i = 0; i < obj.fields.length; i++) pbf.writeMessage(13, FeatureCollectionPBuffer.Field.write, obj.fields[i]);
        if (obj.values) for (i = 0; i < obj.values.length; i++) pbf.writeMessage(14, FeatureCollectionPBuffer.Value.write, obj.values[i]);
        if (obj.features) for (i = 0; i < obj.features.length; i++) pbf.writeMessage(15, FeatureCollectionPBuffer.Feature.write, obj.features[i]);
    };

    // FeatureCollectionPBuffer.CountResult ========================================

    FeatureCollectionPBuffer.CountResult = {};

    FeatureCollectionPBuffer.CountResult.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.CountResult._readField, {count: 0}, end);
    };
    FeatureCollectionPBuffer.CountResult._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.count = pbf.readVarint();
    };
    FeatureCollectionPBuffer.CountResult.write = function (obj, pbf) {
        if (obj.count) pbf.writeVarintField(1, obj.count);
    };

    // FeatureCollectionPBuffer.ObjectIdsResult ========================================

    FeatureCollectionPBuffer.ObjectIdsResult = {};

    FeatureCollectionPBuffer.ObjectIdsResult.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.ObjectIdsResult._readField, {objectIdFieldName: "", serverGens: null, objectIds: []}, end);
    };
    FeatureCollectionPBuffer.ObjectIdsResult._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.objectIdFieldName = pbf.readString();
        else if (tag === 2) obj.serverGens = FeatureCollectionPBuffer.ServerGens.read(pbf, pbf.readVarint() + pbf.pos);
        else if (tag === 3) pbf.readPackedVarint(obj.objectIds);
    };
    FeatureCollectionPBuffer.ObjectIdsResult.write = function (obj, pbf) {
        if (obj.objectIdFieldName) pbf.writeStringField(1, obj.objectIdFieldName);
        if (obj.serverGens) pbf.writeMessage(2, FeatureCollectionPBuffer.ServerGens.write, obj.serverGens);
        if (obj.objectIds) pbf.writePackedVarint(3, obj.objectIds);
    };

    // FeatureCollectionPBuffer.QueryResult ========================================

    FeatureCollectionPBuffer.QueryResult = {};

    FeatureCollectionPBuffer.QueryResult.read = function (pbf, end) {
        return pbf.readFields(FeatureCollectionPBuffer.QueryResult._readField, {featureResult: null, Results: null, countResult: null, idsResult: null}, end);
    };
    FeatureCollectionPBuffer.QueryResult._readField = function (tag, obj, pbf) {
        if (tag === 1) obj.featureResult = FeatureCollectionPBuffer.FeatureResult.read(pbf, pbf.readVarint() + pbf.pos), obj.Results = "featureResult";
        else if (tag === 2) obj.countResult = FeatureCollectionPBuffer.CountResult.read(pbf, pbf.readVarint() + pbf.pos), obj.Results = "countResult";
        else if (tag === 3) obj.idsResult = FeatureCollectionPBuffer.ObjectIdsResult.read(pbf, pbf.readVarint() + pbf.pos), obj.Results = "idsResult";
    };
    FeatureCollectionPBuffer.QueryResult.write = function (obj, pbf) {
        if (obj.featureResult) pbf.writeMessage(1, FeatureCollectionPBuffer.FeatureResult.write, obj.featureResult);
        if (obj.countResult) pbf.writeMessage(2, FeatureCollectionPBuffer.CountResult.write, obj.countResult);
        if (obj.idsResult) pbf.writeMessage(3, FeatureCollectionPBuffer.ObjectIdsResult.write, obj.idsResult);
    };

    function decode(featureCollectionBuffer) {
      let decodedObject;
      try {
        decodedObject = FeatureCollectionPBuffer_1.read(new Pbf__default['default'](featureCollectionBuffer));
      } catch (error) {
        throw new Error('Could not parse arcgis-pbf buffer')
      }
      const featureResult = decodedObject.queryResult.featureResult;
      const transform = featureResult.transform;
      const geometryType = featureResult.geometryType;
      const objectIdField = featureResult.objectIdFieldName;

      // Wires up the field keynames
      const fields = featureResult.fields;
      for (let index = 0; index < fields.length; index++) {
        const field = fields[index];
        field.keyName = getKeyName(field);
      }

      const out = {
        type: 'FeatureCollection',
        features: []
      };

      const geometryParser = getGeometryParser(geometryType);

      const featureLen = featureResult.features.length;
      for (let index = 0; index < featureLen; index++) {
        const f = featureResult.features[index];
        out.features.push({
          type: 'Feature',
          id: getFeatureId(fields, f.attributes, objectIdField),
          properties: collectAttributes(fields, f.attributes),
          geometry: geometryParser(f, transform)
        });
      }

      return {
        featureCollection: out,
        exceededTransferLimit: featureResult.exceededTransferLimit
      }
    }

    // * @property {number} esriGeometryTypePoint=0 esriGeometryTypePoint value
    // * @property {number} esriGeometryTypeMultipoint=1 esriGeometryTypeMultipoint value
    // * @property {number} esriGeometryTypePolyline=2 esriGeometryTypePolyline value
    // * @property {number} esriGeometryTypePolygon=3 esriGeometryTypePolygon value
    // * @property {number} esriGeometryTypeMultipatch=4 esriGeometryTypeMultipatch value
    // * @property {number} esriGeometryTypeNone=127 esriGeometryTypeNone value
    function getGeometryParser (featureType) {
      switch(featureType) {
        case 3:
          return createPolygon
        case 2:
          return createLine
        case 0:
          return createPoint
        default:
          return createPolygon
      }
    }

    function createPoint (f, transform) {
      const p = {
        type: 'Point',
        coordinates: transformTuple(f.geometry.coords, transform)
      };
      return p
    }

    function createLine (f, transform) {
      const p = {
        type: 'LineString',
        coordinates: createLinearRing(f.geometry.coords, transform, 0, f.geometry.lengths[0] * 2)
      };
      return p
    }

    function createPolygon (f, transform) {
      const lengths = f.geometry.lengths.length;

      const p = {
        type: 'Polygon',
        coordinates: []
      };

      if (lengths === 1) {
        p.coordinates.push(createLinearRing(f.geometry.coords, transform, 0, f.geometry.lengths[0] * 2));
      } else {
        p.type = 'MultiPolygon';

        let startPoint = 0;
        for (let index = 0; index < lengths; index++) {
          const stopPoint = startPoint + (f.geometry.lengths[index] * 2);
          const ring = createLinearRing(f.geometry.coords, transform, startPoint, stopPoint);
          
          // Check if the ring is clockwise, if so it's an outer ring
          // If it's counter-clockwise its a hole and so push it to the prev outer ring
          // This is perhaps a bit naive
          // see https://github.com/terraformer-js/terraformer/blob/master/packages/arcgis/src/geojson.js
          // for a fuller example of doing this 
          if (ringIsClockwise(ring)) {
            p.coordinates.push([ring]);
          } else if (p.coordinates.length > 0) {
             p.coordinates[p.coordinates.length - 1].push(ring);
          }
          startPoint = stopPoint;
        }
      }
      
      return p
    }

    function ringIsClockwise (ringToTest) {
      var total = 0;
      var i = 0;
      var rLength = ringToTest.length;
      var pt1 = ringToTest[i];
      var pt2;
      for (i; i < rLength - 1; i++) {
        pt2 = ringToTest[i + 1];
        total += (pt2[0] - pt1[0]) * (pt2[1] + pt1[1]);
        pt1 = pt2;
      }
      return (total >= 0)
    }

    function createLinearRing (arr, transform, startPoint, stopPoint) {
      const out = [];
      if (arr.length === 0) return out


      const initialX = arr[startPoint];
      const initialY = arr[startPoint + 1];
      out.push(transformTuple([initialX, initialY], transform));
      let prevX = initialX;
      let prevY = initialY;
      for (let i = startPoint + 2; i < stopPoint; i = i + 2) {
        const x = difference(prevX, arr[i]);
        const y = difference(prevY, arr[i + 1]);
        const transformed = transformTuple([x, y], transform);
        out.push(transformed);
        prevX = x;
        prevY = y;
      }
      return out
    }

    function collectAttributes(fields, featureAttributes) {
      const out = {};
      for (let i = 0; i < fields.length; i++) {
        const f = fields[i];
        if (featureAttributes[i][featureAttributes[i].value_type]) out[f.name] = featureAttributes[i][featureAttributes[i].value_type];
        else out[f.name] = null;
      }
      return out
    }

    function getFeatureId(fields, featureAttributes, featureIdField) {
      for (let index = 0; index < fields.length; index++) {
        const field = fields[index];
        if (field.name === featureIdField) {
          return featureAttributes[index][featureAttributes[index].value_type]
        }
      }
      return null
    }

    function getKeyName (fields) {
      switch (fields.fieldType) {
      case 1:
        return 'sintValue'
      case 2:
        return 'floatValue'
      case 3:
        return 'doubleValue'
      case 4:
        return 'stringValue'
      case 5:
        return 'sint64Value'
      case 6:
        return 'uintValue'
      default:
        return null
      }
    }

    function transformTuple(coords, transform) {

      let x = coords[0];
      let y = coords[1];

      let z = coords[2] ? coords[2] : undefined;
      if (transform.scale) {
        x *= transform.scale.xScale;
        y *= -transform.scale.yScale;
        if (undefined !== z) { z *= transform.scale.zScale; }
      }
      if (transform.translate) {
        x += transform.translate.xTranslate;
        y += transform.translate.yTranslate;
        if (undefined !== z) { z += transform.translate.zTranslate; }
      }
      const ret = [x, y];
      if (undefined !== z) { ret.push(z); }
      return ret;
    }

    function difference(a, b) {
      return a + b
    }

    class FeatureService {
      constructor(sourceId, map, arcgisOptions, geojsonSourceOptions) {
        if (!sourceId || !map || !arcgisOptions) throw new Error('Source id, map and arcgisOptions must be supplied as the first three arguments.')
        if (!arcgisOptions.url) throw new Error('A url must be supplied as part of the esriServiceOptions object.')

        this.sourceId = sourceId;
        this._map = map;

        this._tileIndices = new Map();
        this._featureIndices = new Map();
        this._featureCollections = new Map();

        this._esriServiceOptions = Object.assign({
          useStaticZoomLevel: false,
          minZoom: arcgisOptions.useStaticZoomLevel ? 7 : 2,
          simplifyFactor: 0.3,
          precision: 8,
          where: '1=1',
          to: null,
          from: null,
          outFields: '*',
          setAttributionFromService: true,
          f: 'pbf',
          useSeviceBounds: true,
          projectionEndpoint: `${arcgisOptions.url.split('rest/services')[0]}rest/services/Geometry/GeometryServer/project`,
          token: null,
          fetchOptions: null
        }, arcgisOptions);

        this._fallbackProjectionEndpoint = 'https://tasks.arcgisonline.com/arcgis/rest/services/Geometry/GeometryServer/project';
        this.serviceMetadata = null;
        this._maxExtent = [-Infinity, Infinity, -Infinity, Infinity];

        const gjOptions = !geojsonSourceOptions ? {} : geojsonSourceOptions;
        this._map.addSource(sourceId, Object.assign(gjOptions, {
          type: 'geojson',
          data: this._getBlankFc()
        }));

        this._getServiceMetadata()
          .then(() => {
            if (!this.supportsPbf) {
              if (!this.supportsGeojson) {
                this._map.removeSource(sourceId);
                throw new Error('Server does not support PBF or GeoJSON query formats.')
              }
              this._esriServiceOptions.f = 'geojson';
            }

            if (this._esriServiceOptions.useSeviceBounds) {
              const serviceExtent = this.serviceMetadata.extent;
              if (serviceExtent.spatialReference.wkid === 4326) {
                this._setBounds([serviceExtent.xmin, serviceExtent.ymin, serviceExtent.xmax, serviceExtent.ymax]);
                this._clearAndRefreshTiles();
              } else {
                this._projectBounds();
              }
            }

            if (this._esriServiceOptions.outFields !== '*') {
              this._esriServiceOptions.outFields = `${this._esriServiceOptions.outFields},${this.serviceMetadata.uniqueIdField.name}`;
            }

            this._setAttribution();
            this.enableRequests();
          });
      }

      destroySource() {
        this.disableRequests();
        this._map.removeSource(this.sourceId);
      }

      _getBlankFc() {
        return {
          type: 'FeatureCollection',
          features: []
        }
      }

      _setBounds(bounds) {
        this._maxExtent = bounds;
      }

      get supportsGeojson() {
        return this.serviceMetadata.supportedQueryFormats.indexOf('geoJSON') > -1
      }

      get supportsPbf() {
        return this.serviceMetadata.supportedQueryFormats.indexOf('PBF') > -1
      }

      disableRequests() {
        this._map.off('moveend', this._boundEvent);
      }

      enableRequests() {
        this._boundEvent = this._findAndMapData.bind(this);
        this._map.on('moveend', this._boundEvent);
      }

      _clearAndRefreshTiles() {
        this._tileIndices = new Map();
        this._featureIndices = new Map();
        this._featureCollections = new Map();
        this._findAndMapData();
      }

      setWhere(newWhere) {
        this._esriServiceOptions.where = newWhere;
        this._clearAndRefreshTiles();
      }

      clearWhere() {
        this._esriServiceOptions.where = '1=1';
        this._clearAndRefreshTiles();
      }

      setDate(to, from) {
        this._esriServiceOptions.to = to;
        this._esriServiceOptions.from = from;
        this._clearAndRefreshTiles();
      }

      setToken(token) {
        this._esriServiceOptions.token = token;
        this._clearAndRefreshTiles();
      }

      _createOrGetTileIndex(zoomLevel) {
        const existingZoomIndex = this._tileIndices.get(zoomLevel);
        if (existingZoomIndex) return existingZoomIndex
        const newIndex = new Map();
        this._tileIndices.set(zoomLevel, newIndex);
        return newIndex
      }

      _createOrGetFeatureCollection(zoomLevel) {
        const existingZoomIndex = this._featureCollections.get(zoomLevel);
        if (existingZoomIndex) return existingZoomIndex
        const fc = this._getBlankFc();
        this._featureCollections.set(zoomLevel, fc);
        return fc
      }

      _createOrGetFeatureIdIndex(zoomLevel) {
        const existingFeatureIdIndex = this._featureIndices.get(zoomLevel);
        if (existingFeatureIdIndex) return existingFeatureIdIndex
        const newFeatureIdIndex = new Map();
        this._featureIndices.set(zoomLevel, newFeatureIdIndex);
        return newFeatureIdIndex
      }

      async _findAndMapData() {
        const z = this._map.getZoom();

        if (z < this._esriServiceOptions.minZoom) {
          return
        }
        const bounds = this._map.getBounds().toArray();
        const primaryTile = tilebelt.bboxToTile([bounds[0][0], bounds[0][1], bounds[1][0], bounds[1][1]]);

        if (this._esriServiceOptions.useSeviceBounds && !this._doesTileOverlapBbox(this._maxExtent, bounds)) {
          return
        }

        // If we're not using a static zoom level we'll round to the nearest even zoom level
        // This means we don't need to request new data for every zoom level allowing us to reuse the previous levels data
        const zoomLevel = this._esriServiceOptions.useStaticZoomLevel ? this._esriServiceOptions.minZoom : 2 * Math.floor(z / 2);
        const zoomLevelIndex = this._createOrGetTileIndex(zoomLevel);
        const featureIdIndex = this._createOrGetFeatureIdIndex(zoomLevel);
        const fc = this._createOrGetFeatureCollection(zoomLevel);

        const tilesToRequest = [];

        if (primaryTile[2] < zoomLevel) {
          let candidateTiles = tilebelt.getChildren(primaryTile);
          let minZoomOfCandidates = candidateTiles[0][2];
          while (minZoomOfCandidates < zoomLevel) {
            const newCandidateTiles = [];
            candidateTiles.forEach(t => newCandidateTiles.push(...tilebelt.getChildren(t)));
            candidateTiles = newCandidateTiles;
            minZoomOfCandidates = candidateTiles[0][2];
          }

          for (let index = 0; index < candidateTiles.length; index++) {
            if (this._doesTileOverlapBbox(candidateTiles[index], bounds)) {
              tilesToRequest.push(candidateTiles[index]);
            }
          }
        } else {
          tilesToRequest.push(primaryTile);
        }

        for (let index = 0; index < tilesToRequest.length; index++) {
          const quadKey = tilebelt.tileToQuadkey(tilesToRequest[index]);
          if (zoomLevelIndex.has(quadKey)) {
            tilesToRequest.splice(index, 1);
            index--;
          } else zoomLevelIndex.set(quadKey, true);
        }

        if (tilesToRequest.length === 0) {
          this._updateFcOnMap(fc);
          return
        }

        // This tolerance will be used to inform the quantization/simplification of features
        const mapWidth = Math.abs(bounds[1][0] - bounds[0][0]);
        const tolerance = (mapWidth / this._map.getCanvas().width) * this._esriServiceOptions.simplifyFactor;
        await this._loadTiles(tilesToRequest, tolerance, featureIdIndex, fc);
        this._updateFcOnMap(fc);
      }

      async _loadTiles(tilesToRequest, tolerance, featureIdIndex, fc) {
        return new Promise((resolve) => {
          const promises = tilesToRequest.map(t => this._getTile(t, tolerance));
          Promise.all(promises).then((featureCollections) => {
            featureCollections.forEach((tileFc) => {
              if (tileFc) this._iterateItems(tileFc, featureIdIndex, fc);
            });
            resolve();
          });
        })
      }

      _iterateItems(tileFc, featureIdIndex, fc) {
        tileFc.features.forEach((feature) => {
          if (!featureIdIndex.has(feature.id)) {
            fc.features.push(feature);
            featureIdIndex.set(feature.id);
          }
        });
      }

      get _time() {
        if (!this._esriServiceOptions.to) return false
        let from = this._esriServiceOptions.from;
        let to = this._esriServiceOptions.to;
        if (from instanceof Date) from = from.valueOf();
        if (to instanceof Date) to = to.valueOf();

        return `${from},${to}`
      }

      _getTile(tile, tolerance) {
        const tileBounds = tilebelt.tileToBBOX(tile);

        const extent = {
          spatialReference: {
            latestWkid: 4326,
            wkid: 4326
          },
          xmin: tileBounds[0],
          ymin: tileBounds[1],
          xmax: tileBounds[2],
          ymax: tileBounds[3]
        };

        const params = new URLSearchParams({
          f: this._esriServiceOptions.f,
          geometry: JSON.stringify(extent),
          where: this._esriServiceOptions.where,
          outFields: this._esriServiceOptions.outFields,
          outSR: 4326,
          returnZ: false,
          returnM: false,
          precision: this._esriServiceOptions.precision,
          quantizationParameters: JSON.stringify({
            extent,
            tolerance,
            mode: 'view'
          }),
          resultType: 'tile',
          spatialRel: 'esriSpatialRelIntersects',
          geometryType: 'esriGeometryEnvelope',
          inSR: 4326
        });

        if (this._time) params.append('time', this._time);

        this._appendTokenIfExists(params);

        return new Promise((resolve) => {
          fetch(`${`${this._esriServiceOptions.url}/query?${params.toString()}`}`, this._esriServiceOptions.fetchOptions)
            .then(response => (this._esriServiceOptions.f === 'pbf' ? response.arrayBuffer() : response.json()))
            .then((data) => {
              let out;
              try {
                out = this._esriServiceOptions.f === 'pbf' ? decode(new Uint8Array(data)).featureCollection : data;
              } catch (err) {
                console.error('Could not parse arcgis buffer. Please check the url you requested.');
              }
              resolve(out);
            });
        })
      }

      _updateFcOnMap(fc) {
        this._map.getSource(this.sourceId).setData(fc);
      }

      _doesTileOverlapBbox(tile, bbox) {
        const tileBounds = tile.length === 4 ? tile : tilebelt.tileToBBOX(tile);
        if (tileBounds[2] < bbox[0][0]) return false
        if (tileBounds[0] > bbox[1][0]) return false
        if (tileBounds[3] < bbox[0][1]) return false
        if (tileBounds[1] > bbox[1][1]) return false
        return true
      }

      _getServiceMetadata() {
        if (this.serviceMetadata !== null) return Promise.resolve(this.serviceMetadata)

        const params = new URLSearchParams({f: 'json'});

        this._appendTokenIfExists(params);
        return this._requestJson(`${this._esriServiceOptions.url}?${params.toString()}`, this._esriServiceOptions.fetchOptions)
          .then((data) => {
            // Esri sends error responses with a 200 status code, so handle them in `.then`
            if (data.error) {
              throw new Error(JSON.stringify(data.error))
            }
            this.serviceMetadata = data;
            return (this.serviceMetadata)
          })
      }

      getFeaturesByLonLat(lnglat, radius, returnGeometry) {
        returnGeometry = returnGeometry ? returnGeometry : false;
        radius = radius ? radius : 20;

        const params = new URLSearchParams({
          sr: 4326,
          geometryType: 'esriGeometryPoint',
          geometry: JSON.stringify({
            x: lnglat.lng,
            y: lnglat.lat,
            spatialReference: {
              wkid: 4326
            }
          }),
          returnGeometry,
          time: this._time,
          outFields: '*',
          spatialRel: 'esriSpatialRelIntersects',
          units: 'esriSRUnit_Meter',
          distance: radius,
          f: 'geojson'
        });

        this._appendTokenIfExists(params);

        return new Promise((resolve) => {
          this._requestJson(`${this._esriServiceOptions.url}/query?${params.toString()}`, this._esriServiceOptions.fetchOptions)
            .then(data => resolve(data));
        })
      }

      getFeaturesByObjectIds(objectIds, returnGeometry) {
        if (Array.isArray(objectIds)) objectIds = objectIds.join(',');
        returnGeometry = returnGeometry ? returnGeometry : false;
        const params = new URLSearchParams({
          sr: 4326,
          objectIds,
          returnGeometry,
          outFields: '*',
          f: 'geojson'
        });

        this._appendTokenIfExists(params);

        return new Promise((resolve) => {
          this._requestJson(`${this._esriServiceOptions.url}/query?${params.toString()}`, this._esriServiceOptions.fetchOptions)
            .then(data => resolve(data));
        })
      }

      _projectBounds() {
        const params = new URLSearchParams({
          geometries: JSON.stringify({
            geometryType: 'esriGeometryEnvelope',
            geometries: [this.serviceMetadata.extent]
          }),
          inSR: this.serviceMetadata.extent.spatialReference.wkid,
          outSR: 4326,
          f: 'json'
        });
        let fetchOptions = {};
        if (!this._projectionEndpointIsFallback()) {
          fetchOptions = this._esriServiceOptions.fetchOptions;
          this._appendTokenIfExists(params);
        }

        this._requestJson(`${this._esriServiceOptions.projectionEndpoint}?${params.toString()}`, fetchOptions)
          .then((data) => {
            const extent = data.geometries[0];
            this._maxExtent = [extent.xmin, extent.ymin, extent.xmax, extent.ymax];
            this._clearAndRefreshTiles();
          })
          .catch((error) => {
            // if projection endpoint has already been set to fallback, do not re-request project bounds
            if (this._projectionEndpointIsFallback()) {
              throw error
            } else {
              this._esriServiceOptions.projectionEndpoint = this._fallbackProjectionEndpoint;
              this._projectBounds();
            }
          });
      }

      _requestJson(url, fetchOptions) {
        return new Promise((resolve, reject) => {
          fetch(url, fetchOptions)
            .then(response => response.json())
            .then((data) => {
              if ('error' in data) reject(new Error('Endpoint doesnt exist'));
              resolve(data);
            })
            .catch(error => reject(error));
        })
      }

      _projectionEndpointIsFallback() {
        return this._esriServiceOptions.projectionEndpoint === this._fallbackProjectionEndpoint
      }

      _setAttribution() {
        const POWERED_BY_ESRI_ATTRIBUTION_STRING = 'Powered by <a href="https://www.esri.com">Esri</a>';

        const attributionController = this._map._controls.find(c => '_attribHTML' in c);

        if (!attributionController) return

        const customAttribution = attributionController.options.customAttribution;

        if (typeof customAttribution === 'string') {
          attributionController.options.customAttribution = `${customAttribution} | ${POWERED_BY_ESRI_ATTRIBUTION_STRING}`;
        } else if (customAttribution === undefined) {
          attributionController.options.customAttribution = POWERED_BY_ESRI_ATTRIBUTION_STRING;
        } else if (Array.isArray(customAttribution)) {
          if (customAttribution.indexOf(POWERED_BY_ESRI_ATTRIBUTION_STRING) === -1) {
            customAttribution.push(POWERED_BY_ESRI_ATTRIBUTION_STRING);
          }
        }

        if (this._esriServiceOptions.setAttributionFromService && this.serviceMetadata.copyrightText.length > 0) {
          this._map.style.sourceCaches[this.sourceId]._source.attribution = this.serviceMetadata.copyrightText;
        }

        attributionController._updateAttributions();
      }

      _appendTokenIfExists(params) {
        const token = this._esriServiceOptions.token;
        if (token !== null) {
          params.append('token', token);
        }
      }
    }

    return FeatureService;

})));
